# coding=utf-8
"""
Data management for BSC-Earth conventions

Focused on working with observations and reconstructions as well as with downloaded
but no cmorized models (like ECMWF System 4)
"""
import os

from bscearth.utils.log import Log

from earthdiagnostics.datamanager import DataManager
from earthdiagnostics.frequency import Frequencies
from earthdiagnostics.variable import VariableType


class ObsReconManager(DataManager):
    """
    Data manager class for CMORized experiments

    Parameters
    ----------
    config: Config
    """

    def __init__(self, config):
        super(ObsReconManager, self).__init__(config)
        data_folders = self.config.data_dir.split(':')
        self.config.data_dir = None
        for data_folder in data_folders:
            if os.path.isdir(os.path.join(data_folder, self.config.data_type,  self.experiment.institute.lower(),
                                          self.experiment.model.lower())):
                self.config.data_dir = data_folder
                break

        if not self.config.data_dir:
            raise Exception('Can not find model data')

        if self.config.data_type in ('obs', 'recon') and self.experiment.chunk_size != 1:
            raise Exception('For obs and recon data chunk_size must be always 1')

    def get_file_path(self, startdate, domain, var, frequency, vartype,
                      box=None, grid=None):
        """
        Return  the path to a concrete file

        :param startdate: file's startdate
        :type startdate: str
        :param domain: file's domain
        :type domain: str
        :param var: file's var
        :type var: str
        :param frequency: file's frequency
        :type frequency: Frequency
        :param box: file's box
        :type box: Box
        :param grid: file's grid
        :type grid: str
        :return: path to the file
        :rtype: str
        :param vartype: Variable type (mean, statistic)
        :type vartype: VariableType
        """
        if frequency is None:
            frequency = self.config.frequency

        folder_path = self._get_folder_path(frequency, domain, var, grid, vartype)
        file_name = self._get_file_name(var, startdate, frequency)

        filepath = os.path.join(folder_path, file_name)
        return filepath

    def _get_folder_path(self, frequency, domain, variable, grid, vartype):
        var_folder = self.get_varfolder(domain, variable, grid, frequency)
        folder_path = os.path.join(self.config.data_dir, self.config.data_type,
                                   self.experiment.institute.lower(),
                                   self.experiment.model.lower(),
                                   frequency.folder_name(vartype),
                                   var_folder)
        return folder_path

    def _get_file_name(self, var, startdate, frequency):
        if startdate:
            if self.config.data_type != 'exp' and frequency != Frequencies.weekly:
                startdate = startdate[0:6]
            return '{0}_{1}.nc'.format(var, startdate)
        else:
            return '{0}.nc'.format(var)

    def request_chunk(self, domain, var, startdate, member, chunk, grid=None, box=None, frequency=None,
                      vartype=VariableType.MEAN):
        """
        Request a given file from the CMOR repository to the scratch folder and returns the path to the scratch's copy

        Parameters
        ----------
        domain: ModelingRealm
        var: str
        startdate: str
        member: int
        chunk: int
        grid: str or None
        box: Box or None
        frequency: Frequency or None
        vartype: VariableType or None

        Returns
        -------
        DataFile

        """
        var = self._get_final_var_name(box, var)
        filepath = self.get_file_path(startdate, domain, var, frequency, vartype, box, grid)
        Log.debug('{0} requested', filepath)
        return self._get_file_from_storage(filepath)

    # noinspection PyUnusedLocal
    def declare_chunk(self, domain, var, startdate, member, chunk, grid=None, region=None, box=None, frequency=None,
                      vartype=VariableType.MEAN, diagnostic=None):
        """
        Declare a variable chunk to be generated by a diagnostic

        Parameters
        ----------
        domain: ModelingRealm
        var: str
        startdate: str
        member: int
        chunk: int
        grid: str or None, optional
        region: Basin or None, optional
        box: Box or None, optional
        frequency: Frequency or None, optional
        vartype: VariableType, optional
        diagnostic: Diagnostic, optional

        Returns
        -------
        DataFile
        """
        if frequency is None:
            frequency = self.config.frequency
        original_name = var
        cmor_var = self.variable_list.get_variable(var)
        if cmor_var:
            var = cmor_var.short_name
        final_name = self._get_final_var_name(box, var)

        filepath = self.get_file_path(startdate, domain, final_name, frequency, vartype, box, grid)
        netcdf_file = self._declare_generated_file(filepath, domain, final_name, cmor_var, self.config.data_convention,
                                                   region, diagnostic, grid, vartype, original_name)
        netcdf_file.frequency = frequency
        Log.debug('{0} will be generated', filepath)
        return netcdf_file
