# coding=utf-8
"""Compute an Atlantic MOC index"""
import netCDF4
import numpy as np
from bscearth.utils.log import Log

from earthdiagnostics.box import Box
from earthdiagnostics.constants import Basins
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticBasinOption, DiagnosticFloatOption
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.utils import Utils, TempFile
from earthdiagnostics.variable import VariableType


class MaxMoc(Diagnostic):
    """
    Compute an Atlantic MOC index

    It finds the maximum  of the annual
    mean meridional overturning in a latitude / depth region

    :original author: Virginie Guemas <virginie.guemas@bsc.es>
    :contributor: Javier Vegas-Regidor<javier.vegas@bsc.es>

    :created: March 2012
    :last modified: June 2016
    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param year: year to compute
    :type year: int
    :param basin: basin to compute
    :type basin: Basin
    :param box: box to compute
    :type box: Box
    """

    alias = 'mocmax'
    "Diagnostic alias for the configuration file"

    vsftmyz = 'vsftmyz'

    def __init__(self, data_manager, startdate, member, year, basin, box):
        Diagnostic.__init__(self, data_manager)
        self.basin = basin
        self.startdate = startdate
        self.member = member
        self.year = year
        self.box = box

    def __str__(self):
        return 'Max moc Startdate: {0} Member: {1} Year: {2} Box: {3} ' \
               'Basin: {4}'.format(self.startdate, self.member, self.year, self.box, self.basin.name)

    def __eq__(self, other):
        if self._different_type(other):
            return False
        return self.startdate == other.startdate and self.member == other.member and self.year == other.year and \
            self.box == other.box and self.basin == other.basin

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Create a job for each complete year to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: minimum latitude, maximum latitude, minimum depth, maximum depth, basin=global
        :type options: list[str]
        :return:
        """
        options_available = (DiagnosticFloatOption('min_lat'),
                             DiagnosticFloatOption('max_lat'),
                             DiagnosticFloatOption('min_depth'),
                             DiagnosticFloatOption('max_depth'),
                             DiagnosticBasinOption('basin', Basins().Global))
        options = cls.process_options(options, options_available)
        box = Box(True)
        box.min_lat = options['min_lat']
        box.max_lat = options['max_lat']
        box.min_depth = options['min_depth']
        box.max_depth = options['max_depth']

        job_list = list()
        for startdate in diags.config.experiment.startdates:
            for member in diags.config.experiment.members:
                years = diags.config.experiment.get_full_years(startdate)
                if len(years) == 0:
                    Log.user_warning('No complete years are available with the given configuration. '
                                     'MaxMoc can not be computed')
                for year in years:
                    job_list.append(MaxMoc(diags.data_manager, startdate, member, year, options['basin'], box))
        return job_list

    def request_data(self):
        """Request data required by the diagnostic"""
        self.variable_file = self.request_year(ModelingRealms.ocean, MaxMoc.vsftmyz,
                                               self.startdate, self.member, self.year)

    def declare_data_generated(self):
        """Declare data to be generated by the diagnostic"""
        self.results = {'vsftmyzmax': self.declare_year(ModelingRealms.ocean, 'vsftmyzmax', self.startdate, self.member,
                                                        self.year, box=self.box, vartype=VariableType.STATISTIC),
                        'vsftmyzmaxlev': self.declare_year(ModelingRealms.ocean, 'vsftmyzmaxlev', self.startdate,
                                                           self.member, self.year, box=self.box,
                                                           vartype=VariableType.STATISTIC),
                        'vsftmyzmaxlat': self.declare_year(ModelingRealms.ocean, 'vsftmyzmaxlat', self.startdate,
                                                           self.member, self.year, box=self.box,
                                                           vartype=VariableType.STATISTIC),
                        'vsftmyzmin': self.declare_year(ModelingRealms.ocean, 'vsftmyzmin', self.startdate, self.member,
                                                        self.year, box=self.box, vartype=VariableType.STATISTIC),
                        'vsftmyzminlev': self.declare_year(ModelingRealms.ocean, 'vsftmyzminlev', self.startdate,
                                                           self.member, self.year, box=self.box,
                                                           vartype=VariableType.STATISTIC),
                        'vsftmyzminlat': self.declare_year(ModelingRealms.ocean, 'vsftmyzminlat', self.startdate,
                                                           self.member, self.year, box=self.box,
                                                           vartype=VariableType.STATISTIC)}

    def compute(self):
        """Run the diagnostic"""
        nco = Utils.nco

        temp = TempFile.get()
        Utils.copy_file(self.variable_file.local_file, temp)

        handler = Utils.open_cdf(temp)
        if 'i' in handler.dimensions:
            handler.close()
            nco.ncwa(input=temp, output=temp, options=('-O -a i',))
        else:
            handler.close()
        handler = Utils.open_cdf(temp)
        basin_index = np.where(handler.variables['basin'][:] == self.basin.name)
        if len(basin_index) == 0:
            raise Exception("Basin {1} is not defined in {0}", temp, self.basin.name)
        basin_index = basin_index[0][0]

        lev = handler.variables['lev'][:]
        if "lat" in handler.variables:
            lat = handler.variables['lat'][:]
        else:
            lat = handler.variables['latitude'][:]

        if self.box.min_lat == self.box.max_lat:
            lat_inds = ((np.abs(lat - self.box.min_lat)).argmin(),)
        else:
            lat_inds = np.where((lat > self.box.min_lat) & (lat < self.box.max_lat))[0]

        if self.box.min_depth == self.box.max_depth:
            lev_inds = ((np.abs(lev - self.box.min_depth)).argmin(),)
        else:
            lev_inds = np.where((lev > self.box.min_depth) & (lev < self.box.max_depth))[0]

        Log.info('Computing year {0}', str(self.year))
        moc = handler.variables['vsftmyz'][:, lev_inds, lat_inds, basin_index]
        handler.close()

        moc = np.mean(moc, 0)

        maximum = np.amax(moc)
        max_index = np.unravel_index(np.argmax(moc), moc.shape)
        # noinspection PyUnresolvedReferences
        max_lev = lev[lev_inds[max_index[0]]]
        # noinspection PyUnresolvedReferences
        max_lat = lat[lat_inds[max_index[1]]]

        minimum = np.amin(moc)
        minimum_index = np.unravel_index(np.argmin(moc), moc.shape)
        # noinspection PyUnresolvedReferences
        min_lev = lev[lev_inds[minimum_index[0]]]
        # noinspection PyUnresolvedReferences
        min_lat = lat[lat_inds[minimum_index[1]]]

        Log.info('Maximum {0} Sv, latitude {1} depth {2} m', maximum, max_lat, max_lev)
        Log.info('Minimum {0} Sv, latitude {1} depth {2} m', minimum, min_lat, min_lev)

        handler, temp = self._create_output_file()
        var = handler.createVariable('vsftmyzmax', float, ('time',))
        var.long_name = 'Maximum_Overturning'
        var.units = 'Sverdrup'
        var.valid_min = -1000.
        var.valid_max = 1000.
        var[0] = maximum
        handler.close()
        self.results['vsftmyzmax'].set_local_file(temp)

        handler, temp = self._create_output_file()
        var = handler.createVariable('vsftmyzmaxlat', float, ('time',))
        var.long_name = 'Latitude_of_Maximum_Overturning'
        var.units = 'Degrees'
        var.valid_min = -90.
        var.valid_max = 90.
        var[0] = max_lat
        handler.close()
        self.results['vsftmyzmaxlat'].set_local_file(temp)

        handler, temp = self._create_output_file()
        var = handler.createVariable('vsftmyzmaxlev', float, ('time',))
        var.long_name = 'Depth_of_Maximum_Overturning'
        var.units = 'Meters'
        var.valid_min = 0.
        var.valid_max = 10000.
        var[0] = max_lev
        handler.close()
        self.results['vsftmyzmaxlev'].set_local_file(temp)

        handler, temp = self._create_output_file()
        var = handler.createVariable('vsftmyzmin', float, ('time',))
        var.long_name = 'Minimum_Overturning'
        var.units = 'Sverdrup'
        var.valid_min = -1000.
        var.valid_max = 1000.
        var[0] = minimum
        handler.close()
        self.results['vsftmyzmin'].set_local_file(temp)

        handler, temp = self._create_output_file()
        var = handler.createVariable('vsftmyzminlat', float, ('time',))
        var.long_name = 'Latitude_of_Minimum_Overturning'
        var.units = 'Degrees'
        var.valid_min = -90.
        var.valid_max = 90.
        var[0] = min_lat
        handler.close()
        self.results['vsftmyzminlat'].set_local_file(temp)

        handler, temp = self._create_output_file()
        var = handler.createVariable('vsftmyzminlev', float, ('time',))
        var.long_name = 'Depth_of_Minimum_Overturning'
        var.units = 'Meters'
        var.valid_min = 0.
        var.valid_max = 10000.
        var[0] = min_lev
        handler.close()
        self.results['vsftmyzminlev'].set_local_file(temp)

    def _create_output_file(self):
        temp = TempFile.get()
        handler = netCDF4.Dataset(temp, 'w')
        handler.createDimension('time')

        time = handler.createVariable('time', 'i2', ('time',))
        time.calendar = 'gregorian'
        time.units = 'days since January 1, {0}'.format(self.year)
        return handler, temp
