# coding=utf-8
from unittest import TestCase
from mock import Mock
import six
from earthdiagnostics.work_manager import Downloader, WorkManager
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticStatus
from earthdiagnostics.datafile import DataFile, StorageStatus, LocalStatus
from bscearth.utils import log


class TestDownloader(TestCase):

    def setUp(self):
        self.downloader = Downloader()

    def test_start_and_stop(self):
        self.downloader.start()
        self.assertTrue(self.downloader._thread.is_alive())
        self.downloader.shutdown()
        self.assertFalse(self.downloader._thread.is_alive())

    def test_submit(self):
        datafile = Mock()
        self.downloader.submit(datafile)

    def test_download(self):
        datafile = Mock()
        self.downloader.submit(datafile)
        self.downloader.start()
        self.downloader.shutdown()
        datafile.download.assert_called_once()

    def test_download_fails(self):
        datafile = Mock()

        def _download():
            raise Exception('Error')
        datafile.download = _download
        self.downloader.submit(datafile)
        if six.PY3:
            with self.assertLogs(log.Log.log, log.Log.CRITICAL):
                self.downloader.start()
                self.downloader.shutdown()
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_smaller_first(self):
        small_file = self._create_datafile_mock(size=1)
        self.downloader.submit(small_file)
        large_file = self._create_datafile_mock(size=10)
        self.downloader.submit(large_file)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: () Size: 1',
                                  'INFO:bscearth.utils:Suscribers: () Size: 10'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_not_none_first(self):
        small_file = self._create_datafile_mock(size=None)
        self.downloader.submit(small_file)
        large_file = self._create_datafile_mock(size=10)
        self.downloader.submit(large_file)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: () Size: 10',
                                  'INFO:bscearth.utils:Suscribers: () Size: None'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_not_none_second(self):
        small_file = self._create_datafile_mock(size=1)
        self.downloader.submit(small_file)
        large_file = self._create_datafile_mock(size=None)
        self.downloader.submit(large_file)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: () Size: 1',
                                  'INFO:bscearth.utils:Suscribers: () Size: None'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_both_none(self):
        small_file = self._create_datafile_mock(size=None)
        self.downloader.submit(small_file)
        large_file = self._create_datafile_mock(size=None)
        self.downloader.submit(large_file)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: () Size: None',
                                  'INFO:bscearth.utils:Suscribers: () Size: None'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_can_not_order(self):
        small_file = self._create_datafile_mock(size=1)
        self.downloader.submit(small_file)
        large_file = self._create_datafile_mock(size=1)
        self.downloader.submit(large_file)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: () Size: 1',
                                  'INFO:bscearth.utils:Suscribers: () Size: 1'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_more_suscribers_first(self):
        no_suscribers = self._create_datafile_mock()
        self.downloader.submit(no_suscribers)
        suscriber = Mock()
        one_suscriber = self._create_datafile_mock(suscribers=(suscriber, ))
        self.downloader.submit(one_suscriber)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: ({0},) Size: 0'.format(str(suscriber)),
                                  'INFO:bscearth.utils:Suscribers: () Size: 0'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def test_download_more_suscribers_waiting_first(self):

        class StatusDiag(Diagnostic):
            def __init__(self, data_manager, pending):
                super(StatusDiag, self).__init__(data_manager)
                self.pending = pending

            def __str__(self):
                return str(self.pending)

            def pending_requests(self):
                return self.pending

        suscriber = StatusDiag(Mock(), 1)
        suscriber.status = DiagnosticStatus.WAITING
        one_waiting_suscriber = self._create_datafile_mock(suscribers=(suscriber,))
        self.downloader.submit(one_waiting_suscriber)
        suscriber = StatusDiag(Mock(), 2)
        suscriber.status = DiagnosticStatus.FAILED
        one_failed_suscriber = self._create_datafile_mock(suscribers=(suscriber, ))
        self.downloader.submit(one_failed_suscriber)
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.downloader.start()
                self.downloader.shutdown()
            self.assertListEqual(cmd.output,
                                 ['INFO:bscearth.utils:Suscribers: (1,) Size: 0',
                                  'INFO:bscearth.utils:Suscribers: (2,) Size: 0'])
        else:
            self.downloader.start()
            self.downloader.shutdown()

    def _create_datafile_mock(self, size=0, suscribers=tuple()):
        data_mock = Mock()
        data_mock.size = size
        data_mock.suscribers = suscribers

        def _download():
            log.Log.info('Suscribers: {0.suscribers} Size: {0.size}'.format(data_mock))
        data_mock.download = _download

        return data_mock


class MockFile(DataFile):

    def __init__(self):
        super().__init__()
        self.upload_fails = False

    def download(self):
        self.local_status = LocalStatus.READY

    def upload(self):
        if self.upload_fails:
            self.storage_status = StorageStatus.FAILED
        else:
            self.storage_status = StorageStatus.READY

    def clean_local(self):
        pass


class TestWorkManager(TestCase):

    def setUp(self):
        self.config = Mock()
        self.data_manager = Mock()
        self.config.max_cores = 1
        self.config.parallel_uploads = 1
        self.data_manager.requested_files = {}
        self.data_manager.request_chunk = self._request_chunk
        self.data_manager.declare_chunk = self._declare_chunk
        self.work_manager = WorkManager(self.config, self.data_manager)

    def test_prepare_job_list(self):

        class Diag1(Diagnostic):
            alias = 'diag1'

            @classmethod
            def generate_jobs(cls, diags, options):
                cls.process_options(options, [])
                diag = Diag1(self.data_manager)
                diag.add_subjob(Diag1(self.data_manager))
                return (diag,)

        self.config.get_commands.return_value = ['diag1', 'baddiag', 'diag1,badoption']
        Diagnostic.register(Diag1)
        self.work_manager.prepare_job_list()

    def test_run(self):
        self.config.max_cores = -1

        class EmptyDiag(Diagnostic):
            alias = 'diag1'

            def __str__(self):
                return 'Null diag'

            def compute(self):
                pass

            def request_data(self):
                pass

            def declare_data_generated(self):
                pass

        diag = EmptyDiag(self.data_manager)
        self.work_manager.add_job(diag)
        self.work_manager.run()

        self.assertIn(diag, self.work_manager.jobs[DiagnosticStatus.COMPLETED])
        self.assertFalse(self.work_manager.had_errors)

    def test_diag_can_be_skipped(self):
        self.config.skip_diags_done = True

        class SkippedDiag(Diagnostic):
            alias = 'diag1'

            def __str__(self):
                return 'Skipped diag'

            def compute(self):
                raise Exception('Diagnostic should not be computed')

            def request_data(self):
                pass

            def declare_data_generated(self):
                pass

            def can_skip_run(self):
                return True

        diag = SkippedDiag(self.data_manager)
        self.work_manager.add_job(diag)
        self.work_manager.run()

        self.assertIn(diag, self.work_manager.jobs[DiagnosticStatus.COMPLETED])
        self.assertFalse(self.work_manager.had_errors)

    def test_diag_can_be_skipped_but_no(self):
        self.config.skip_diags_done = False

        class SkippedDiag(Diagnostic):
            alias = 'diag1'

            def __str__(self):
                return 'Skipped diag'

            def compute(self):
                raise Exception('Diagnostic should not be computed')

            def request_data(self):
                pass

            def declare_data_generated(self):
                pass

            def can_skip_run(self):
                return True

        diag = SkippedDiag(self.data_manager)
        self.work_manager.add_job(diag)
        self.work_manager.run()

        self.assertIn(diag, self.work_manager.jobs[DiagnosticStatus.FAILED])
        self.assertTrue(self.work_manager.had_errors)

    def test_failed_run(self):
        class FailDiag(Diagnostic):
            alias = 'diag1'

            def __str__(self):
                return 'Fail diag'

            def compute(self):
                raise Exception('Must fail')

            def request_data(self):
                pass

            def declare_data_generated(self):
                pass

            def can_skip_run(self):
                return False

        diag = FailDiag(self.data_manager)
        self.work_manager.add_job(diag)
        self.work_manager.run()

        self.assertIn(diag, self.work_manager.jobs[DiagnosticStatus.FAILED])
        self.assertTrue(self.work_manager.had_errors)

    def _request_chunk(self, *args, **kwargs):
        req = MockFile()
        req.storage_status = StorageStatus.READY
        req.local_status = LocalStatus.PENDING
        req.remote_file = 'requested'
        self.data_manager.requested_files[req.remote_file] = req
        return req

    def _declare_chunk(self, *args, **kwargs):
        req = MockFile()
        req.remote_file = 'generated'
        self.data_manager.requested_files[req.remote_file] = req
        req.storage_status = StorageStatus.PENDING
        req.local_status = LocalStatus.NOT_REQUESTED
        return req

    def test_run_data(self):
        self.data_manager.config.max_cores = -1

        class DataDiag(Diagnostic):
            alias = 'diag1'

            def __str__(self):
                return 'Data diag'

            def compute(self):
                self.declared.local_status = LocalStatus.READY

            def request_data(self):
                self.request = self.request_chunk(None, None, None, None, None)

            def declare_data_generated(self):
                self.declared = self.declare_chunk(None, None, None, None, None)

        diag = DataDiag(self.data_manager)
        self.work_manager.add_job(diag)
        self.work_manager.run()

        self.assertIn(diag, self.work_manager.jobs[DiagnosticStatus.COMPLETED])
        self.assertFalse(self.work_manager.had_errors)

    def test_failed_run_upload_failed(self):

        self.data_manager.config.max_cores = -1

        class DataDiag(Diagnostic):
            alias = 'diag1'

            def __str__(self):
                return 'Data diag'

            def compute(self):
                self.declared.local_status = LocalStatus.READY

            def request_data(self):
                self.request = self.request_chunk(None, None, None, None, None)

            def declare_data_generated(self):
                self.declared = self.declare_chunk(None, None, None, None, None)
                self.declared.upload_fails = True

        data_diag = DataDiag(self.data_manager)
        self.work_manager.add_job(data_diag)
        self.work_manager.run()

        self.assertIn(data_diag, self.work_manager.jobs[DiagnosticStatus.COMPLETED])
        self.assertTrue(self.work_manager.had_errors)

    def test_run_empty(self):
        if six.PY3:
            with self.assertLogs(log.Log.log) as cmd:
                self.work_manager.run()
            self.assertTrue([record for record in cmd.records if
                             record.message == 'No diagnostics to run'])
        else:
            self.work_manager.run()
